/**
  *  \file client/widgets/chartdisplayconfig.cc
  *  \brief WChartDisplayConfig widget
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/chartdisplayconfig.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "client/chart.h"
#include "resmgr/resmgr.h"

namespace {
    struct Option {
        uint16_t    flag;
        const char* name;
    };
}

static const Option flags[] = {
    { GChartOptions::co_Ion,       N_("Ion storms") },
    { GChartOptions::co_Mine,      N_("Minefields") },
    { GChartOptions::co_Ufo,       N_("Ufos") },
    { GChartOptions::co_Sectors,   N_("Sector borders") },
    { GChartOptions::co_Borders,   N_("Starchart borders") },
    { GChartOptions::co_Selection, N_("Selection") },
    { GChartOptions::co_Drawings,  N_("Own drawings") },
    { GChartOptions::co_Labels,    N_("Object labels") },
    { GChartOptions::co_Trails,    N_("Ship trails") },
    { GChartOptions::co_ShipDots,  N_("Ships are dots") },
    { GChartOptions::co_WarpWells, N_("Warp wells") },
    { GChartOptions::co_Messages,  N_("Message flag") },
};

/*
 *  Tree node Ids. A tree node has a 32-bit Id. This folds two pieces of information
 *  into one such value:
 *  - the option index into the above flags[] array
 *  - the sub-index for the actual option:
 *      0 - node
 *      1,2,3 - small/regular/scanner leaf
 */

static inline int32_t
makePair(uint_t index, uint_t sub)
{
    return (index << 3) + sub;
}

static inline uint_t
getIndexFromPair(int32_t pair)
{
    return (pair >> 3);
}

static inline uint_t
getSubIndexFromPair(int32_t pair)
{
    return (pair & 7);
}


/** Constructor.
    \param id Widget id */
WChartDisplayConfig::WChartDisplayConfig(int id)
    : UITreeWidget(id, 16 /* height */, 25*font_ems[FONT_NORMAL])
{
    init();
}

WChartDisplayConfig::~WChartDisplayConfig()
{
}

void
WChartDisplayConfig::drawItem(GfxContext& ctx, GfxRect area, const Item& it)
{
    UITreeWidget::drawItem(ctx, area, it);
    if (area.w >= 20 && area.h >= 20) {
        Value v = getValue(it.id);
        Ptr<GfxPixmap> pix =
            v == Mixed
            ? requestPixmap("cbm")
            : v == Filled
            ? requestPixmap("cbf")
            : v == Inside
            ? requestPixmap("cbi")
            : v == Checked
            ? requestPixmap(ResId("cb", 2))
            : requestPixmap(ResId("cb", 1));
        if (pix) {
            pix->blit(ctx.getCanvas(), area.x + area.w - 18, area.y + area.h/2 - 8);
        }
    }
}

bool
WChartDisplayConfig::handleEvent(const UIEvent& event, bool second_pass)
{
    if (hasState(st_Focused) && event.type == ev_Key && event.key.code == ' ') {
        onClick(getItemId(getCurrentItemIndex()));
        return true;
    } else {
        return UITreeWidget::handleEvent(event, second_pass);
    }
}

void
WChartDisplayConfig::onItemClick(int item, int x)
{
    if (x >= getExtent().w - 20) {
        onClick(getItemId(getCurrentItemIndex()));
    } else {
        UITreeWidget::onItemClick(item, x);
    }
}

void
WChartDisplayConfig::load()
{
    for (int i = 0; i < 3; ++i) {
        opts[i].copyFrom(getChartOpts(GChartOptionIndex(i)));
    }
}

void
WChartDisplayConfig::save()
{
    for (int i = 0; i < 3; ++i) {
        getChartOpts(GChartOptionIndex(i)).copyFrom(opts[i]);
    }
}

void
WChartDisplayConfig::init()
{
    // Content
    for (int i = 0; i < 3; ++i) {
        opts[i].show = opts[i].fill = 0;
    }

    // Configure list
    setItemHeight(std::max(font_heights[FONT_NORMAL] + 4, 20));

    // Configure tree
    for (uint_t i = 0; i < countof(flags); ++i) {
        addItem(makePair(i, 0), 0, _(flags[i].name));
        addItem(makePair(i, NormalChartOptions  + 1), 1, _("Regular starchart"));
        addItem(makePair(i, SmallChartOptions   + 1), 1, _("Small starchart"));
        addItem(makePair(i, ScannerChartOptions + 1), 1, _("Control screen scanner"));
    }
    updateAfterModification();
}

void
WChartDisplayConfig::onClick(int32_t id)
{
    uint_t index = getIndexFromPair(id);
    uint_t sub   = getSubIndexFromPair(id);
    int32_t flag = flags[index].flag;
    if (sub == 0) {
        // Node
        if (getValue(id) == Mixed) {
            // Mixed values: enable all to make them identical
            opts[0].setOption(flag);
        } else {
            // Identical values: toggle
            opts[0].toggleOptions(flag);
        }
        opts[1].copyValues(opts[0], flag);
        opts[2].copyValues(opts[0], flag);
    } else {
        // Single option
        opts[sub-1].toggleOptions(flag);
    }
    drawWidget();
}

WChartDisplayConfig::Value
WChartDisplayConfig::getValue(int32_t id)
{
    uint_t index = getIndexFromPair(id);
    uint_t sub   = getSubIndexFromPair(id);
    int32_t flag = flags[index].flag;

    int32_t show;
    int32_t fill;
    if (sub == 0) {
        // Node. Find out whether the children are different.
        int32_t nshow = (opts[0].show & opts[1].show & opts[2].show) & flag;
        int32_t nfill = (opts[0].fill & opts[1].fill & opts[2].fill) & flag;
        show          = (opts[0].show | opts[1].show | opts[2].show) & flag;
        fill          = (opts[0].fill | opts[1].fill | opts[2].fill) & flag;
        if (show != nshow || fill != nfill) {
            return Mixed;
        }
    } else {
        // Single option.
        const GChartOptions& opt = opts[sub-1];
        show = opt.show;
        fill = opt.fill;
    }

    if ((show & flag) == 0) {
        return Unchecked;
    } else if ((fill & flag) == 0) {
        return Checked;
    } else {
        return (flag == GChartOptions::co_Sectors ? Inside : Filled);
    }
}
