/**
  *  \file game/fcode.h
  *  \brief Friendly Codes
  *
  *  (c) 2001-2021 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_FCODE_H
#define PCC_V2_GAME_FCODE_H

#include <vector>
#include <iosfwd>
#include "cpluslib/string.h"
#include "game/obj.h"
#include "game/playerset.h"

class GPlanet;
class GShip;
class Stream;

//! A friendly code definition.
class GFCode {
 public:
    GFCode();
    GFCode(string_t c, string_t desc_line);
    ~GFCode();

    const string_t& getFCode() const;
    string_t getDescription() const;

    /** Get flags.
        \return bitfield of fc_XXX */
    int getFlags() const
        { return flags; }

    /** Get set of players who can use this FCode. */
    GPlayerSet getRaceMask() const
        { return races; }

    bool worksOn(const GObject& o) const;
    bool worksOn(const GShip& s) const;
    bool worksOn(const GPlanet& p) const;

    enum {
        fc_Ship       = 1,      ///< Works on ships.
        fc_Planet     = 2,      ///< Works on planets.
        fc_Base       = 4,      ///< Works on starbases
        fc_Capital    = 8,      ///< Works on capital ships only.
        fc_Alchemy    = 16,     ///< Works on alchemy ships only.
        fc_Registered = 32,     ///< Works for registered players only.
        fc_Unspecial  = 64      ///< Not a special friendly code per se.
    };

    static bool isNumeric(const string_t& fc);
    static bool isExtraFC(const string_t& fc);
    static bool isSpecial(const string_t& fc, bool ignore_case);
    static bool isUniversalMinefieldFCode(const string_t& fc, bool tolerant);
    static int  getNumericValue(const string_t& fc);

    static bool isAllowedRandom(const string_t& fc);
    static string_t generateRandomFCode();

    static void clearExtraFC();
    static void loadExtraFC(Stream&);

 private:
    void initFromString(const string_t& desc_line);

    string_t      code;
    string_t      description;
    GPlayerSet    races;
    int           flags;
};

class GFCodeList {
    typedef std::vector<GFCode> container_t;
    container_t data;
    
    static GFCodeList instance;
 public:
    typedef container_t::const_iterator  iterator;

    GFCodeList();
    GFCodeList(const GFCodeList&, const GObject& o);
    ~GFCodeList();

    /* container interface */
    container_t::size_type getSize() const
        { return data.size(); }
    iterator begin() const
        { return data.begin(); }
    iterator end() const
        { return data.end(); }

    /* manipulator interface */
    void addFCode(const GFCode& code);
    void sort();
    void clear();

    void loadFromFile(Stream& in);

    /* accessor interface */
    const GFCode& operator[](container_t::size_type n) const
        { return data[n]; }
    iterator getFCodeByName(string_t fc) const;

    static GFCodeList& getInstance()
        { return instance; }
};

extern const struct ModuleInitializer mod_fcode[];

#endif
